@@ -1,5 +1,3 @@
-#!/usr/bin/python
-
 #
 ##
 ### MoThEr. Mommas gonna make all of your nightmares come true.
@@ -49,6 +47,7 @@ SQL_FALSE   = "False"
 
 
 _J= ', '.join
+_A= ' AND '.join
 
 #
 ##
@@ -71,6 +70,7 @@ def init_mother(cfile, pg_conn= None, fnaming= None):
     import speaker
     execfile(cfile, speaker.__dict__, d)
 
+    # Handle OIDs or not?
     use_oids= d.get('MOTHER_OIDS', True)
     if not use_oids:
         from dbda import PsygresFly
@@ -89,11 +89,14 @@ def init_mother(cfile, pg_conn= None, fnaming= None):
     speaker.init_speaker(d)
     init_dbda(d, pg_conn)
 
+    # Load DB map
+    momap= d['MOTHER_MAP']
+    _DbMap._load_map(momap)
+
+    # Design Mother
     DbMother.MotherNaming= staticmethod(fnaming)
     DbMother._mo_cfile= cfile
-    DbMother._mo_dbstruct= d['MOTHER_MAP']
     DbMother._oided_mother= use_oids
-    DbMother._mo_initialized= True
     DbMother.__init__= DbMother.__post_init__
     DbMother._sqlInsert= _sqli
     DbMother._insert= _i
@@ -105,7 +108,6 @@ def init_mother(cfile, pg_conn= None, fnaming= None):
     del DbMother._sqlInsertNoOid
     del DbMother._insertOid
     del DbMother._insertNoOid
-    del DbMother._if_not_initialized
     del DbMother._mo_initialized
 
 
@@ -119,44 +121,93 @@ def MotherSession(name= None, pg_conn= None):
 
     return PsygresPool.newSession(name, pg_conn) 
 
-#
-## don't use this:
 
-def _MotherMap(*args):
 
-    from mother.commons import somethingFromFile
-    try:
-        return somethingFromFile(DbMother._mo_dbstruct, *args)
-    except:
+class _DbMap(Psygres):
+    _map_file=       None
+    _map_fields=     None
+    _map_pkeys=      None
+    _map_children=   None
+    _map_rels=       None
+
+    @staticmethod
+    def _load_map(map_file):
         from speaker import Speaker
-        Speaker.log_int_raise(
-            "Error loading Mother Map; file %s is "
-            "invalid or mothermapper.py has to be launched.", 
-            ERR_COL(DbMother._mo_dbstruct))
+        from os import path
+        if not path.isfile(map_file):
+            Speaker.log_int_raise("Mother Map %s does not exist.", ERR_COL(map_file)) 
+        
+        try:
+            fil= open(map_file)
+        except:
+            Speaker.log_int_raise("Unable to open Mother Map %s.", ERR_COL(map_file)) 
 
+        import cPickle
+        try:
+            map_dicts= cPickle.load(fil)
+            _DbMap._map_file= map_file
+            _DbMap._map_fields= map_dicts['K']
+            _DbMap._map_pkeys= map_dicts['P']
+            _DbMap._map_children= map_dicts['C']
+            _DbMap._map_rels= map_dicts['R']
+        except:
+            try:
+                fil.close()
+            except:
+                pass
 
+            Speaker.log_int_raise("Mother Map %s is malformed.", ERR_COL(map_file)) 
 
-## ... use these:
-#
+        fil.close()
 
-def _MotherMappedFields(): 
-    return _MotherMap('TABLE_FIELDS_DICT')[0]
+    @staticmethod
+    def _table_fields(tbl):
+        return _DbMap._map_fields[tbl]
 
-def _MotherMappedPkeys():
-    return _MotherMap('TABLE_PKEYS_DICT')[0]
+    @staticmethod
+    def _table_pkeys(tbl):
+        return _DbMap._map_pkeys[tbl]
 
-def _MotherMappedTables():
-    return _MotherMap('TABLE_FIELDS_DICT', 'TABLE_PKEYS_DICT')
+    @staticmethod
+    def _table_children(tbl):
+        return _DbMap._map_children[tbl]
 
-def _MotherMappedChildren():
-    return _MotherMap('TABLE_CHILDS_DICT')[0]
+    @staticmethod
+    def _searchRelation(tbls):
 
-def _MotherMappedRels():
-    return _MotherMap('TABLE_RELATIONS_DICT')[0]
+        from speaker import Speaker
+        fz= frozenset(tbls)
 
-def _MotherMappedDeps():
-    return _MotherMap('TABLE_CHILDS_DICT', 'TABLE_RELATIONS_DICT')
+        trd= _DbMap._map_rels
+        try:
+            r_tbl= trd[fz]
+            return r_tbl
 
+        except:
+            pass
+
+        Speaker.log_insane(\
+            "Warning: direct relation for tables %s "
+            "is not defined: searching subset...",
+            INF_COL(list(fz)))
+
+        rels= [trd[k] for k in trd.keys() if fz< k]
+
+        if len(rels)<>1:
+            Speaker.log_int_raise(
+                "In _searchRelation: can not get the "  
+                "relation table. There are %s possible "     
+                "choices: %s .", ERR_COL(len(rels)), 
+                ERR_COL(_J(rels)))
+
+        return rels[0]
+
+    @staticmethod
+    def _isChildOf(c, f):
+        return c in _DbMap._map_children[f]
+
+def _do_nothing(*args):
+    pass
 
 #
 ##
@@ -164,13 +215,12 @@ def _MotherMappedDeps():
 ##
 #
 
-class DbMother(Psygres):
+class DbMother(_DbMap):
     """ The Holy MoThEr 
         An abstract class which represents a db Table.
     """
     _mo_initialized= False
     _mo_cfile= None
-    _mo_dbstruct= None
     _mo_false_values= [SQL_NULL, SQL_DEFAULT]
     _mo_fuck= " MoThEr fucker!"
     _mo_session_name= "MoSession-"
@@ -188,14 +238,11 @@ class DbMother(Psygres):
     def _flag_actions(obj, flag):
         if flag == MO_UP: return obj.update
         if flag == MO_DEL: return obj.delete
+        if flag == MO_NOA: return _do_nothing
         if flag == MO_SAVE: return obj.insert
         if flag == MO_LOAD: return obj.load
         if flag == MO_ULOAD: return obj.uload 
-        if flag == MO_NOA: 
-            def do_nothing(*args):
-                pass
-            return do_nothing
-        obj.log_int_raise(self._mo_fuck)
+        obj.log_int_raise("Invalid flag %s", ERR_COL(flag))
 
     @staticmethod
     def _trigger_actions(obj, flag):
@@ -203,21 +250,13 @@ class DbMother(Psygres):
         if flag == MO_DEL: return obj._delete
         if flag == MO_SAVE: return obj._insert
         if flag == MO_LOAD: return obj._load
-        obj.log_int_raise(self._mo_fuck)
-
-    def _if_not_initialized(self):
-        if not DbMother._mo_initialized:
-            self.log_int_raise("Mother not Initialized yet.")
+        obj.log_int_raise("Invalid flag %s", ERR_COL(flag))
 
     def __init__(self, store= {}, flag= MO_NOA, session= None):
         """ __init__(dict_values, flag, session)
         """
 
-        self._if_not_initialized()
-        # This is an hack to bypass initialization control
-        # after the first Mother initialization
-        self.__post_init__(store, flag, session)
-
+        self.log_int_raise(ERR_COL("Mother not Initialized yet."))
 
     def __post_init__(self, store= {}, flag= MO_NOA, session= None):
         """ __init__(dict_values, flag, session)
@@ -225,7 +264,7 @@ class DbMother(Psygres):
 
         # table_name is needed...
         if not hasattr(self, 'table_name'):
-            self.log_int_raise("No table_name, %s", self._mo_fuck)
+            self.log_int_raise("No table_name defined for this Mother.")
 
         # Useful var
         tbl= self.table_name
@@ -235,20 +274,16 @@ class DbMother(Psygres):
         # You can set these vars in the child class or let
         # MoThEr assign them dynamically.
         if not hasattr(self, "fields"):
-            tfd= _MotherMappedFields()
-            self.fields = tfd[tbl]
-            del tfd
+            self.fields = self._table_fields(tbl)
            
         if not hasattr(self, "pkeys"):
-            tfd= _MotherMappedPkeys()
-            self.pkeys = tfd[tbl]
-            del tfd
+            self.pkeys = self._table_pkeys(tbl)
 
         # Frozeset!
         self.fields= frozenset(self.fields)
         self.pkeys= frozenset(self.pkeys)
 
-        # Use Universe() if specified
+        # Session() ?
         if session:
             session._import(self)
         self.session= session
@@ -287,20 +322,14 @@ class DbMother(Psygres):
             self.log_int_raise("Invalid Fields %s for table |%s|.",\
                                 ERR_COL(list(accio)), self.table_name)
 
-        self._moved=set([])
-        self._store=store.copy()
-
-        # Very Bad Flag. CRASH!
-        if flag not in self._mo_flags:
-            self.log_int_raise("Mth: invalid Flag %s, %s",\
-                ERR_COL(self._mo_flags[flag]), self._mo_fuck)
+        self._moved= set([])
+        self._store= store.copy()
 
         # Signal Mother that passed fields are
         # new fields 
         self._moved= set([k for k in setK if k not in self.pkeys])
 
-        f= self._flag_actions(self, flag)
-        f()
+        self._flag_actions(self, flag)()
 
 #
 ##
@@ -382,12 +411,14 @@ class DbMother(Psygres):
         for k in self.pkeys | params:
             res.append('%s= %%(%s)s' % (k, k))
 
-        res= ' AND '.join(res)
+        res= _A(res)
         return self._sqlMogrify(res)
 
     def _sqlUpdate(self, d, params= None):
+
         res= []
         s= self._store
+
         for k in d:
             v= s[k]
             if v== SQL_DEFAULT:
@@ -438,13 +469,16 @@ class DbMother(Psygres):
         return self._sqlMogrify(qry)
 
     def _sqlDelete(self, params= None):
+
         filter= self._sqlWherePkeys(params)
         return ("DELETE FROM %s WHERE %s" % 
                 (self.table_name, filter))
 
     def _sqlSelect(self, d, filter= None, tbl= None):
+
         if filter is None:
             filter= self._sqlWherePkeys()
+
         tbl= tbl or self.table_name
         what= ['%s.%s' % (tbl, f) for f in d]
 
@@ -476,7 +510,7 @@ class DbMother(Psygres):
                     res.append('%s.%s= DEFAULT' % (tbl, k))
                 else:
                     res.append('%s.%s= %%(%s)s' % (tbl, k, k))
-            s= self.mogrify(' AND '.join(res), filter)
+            s= self.mogrify(_A(res), filter)
 
         elif t(set) or t(frozenset) or t(list):
             res= []
@@ -486,7 +520,7 @@ class DbMother(Psygres):
                     res.append('%s.%s= DEFAULT' % (tbl, k))
                 else:
                     res.append('%s.%s= %%(%s)s' % (tbl, k, k))
-            s=self._sqlMogrify(' AND '.join(res))
+            s=self._sqlMogrify(_A(res))
 
         else:
             self.log_int_raise(\
@@ -549,6 +583,7 @@ class DbMother(Psygres):
         """
         if not self._triggers_are_initialized():
             return False
+
         return len(self.get_flag_triggers(flag)) > 0
 
     def add_trigger(self, flag, when, trigger):
@@ -995,44 +1030,32 @@ class MotherManager:
         """
 
         s= " JOIN %s on " % r_tbl
-        s+="AND".join(\
+        s+=_A(\
             [" %s.%s = %s.%s " % (r_tbl,v,j_tbl,k)
                 for k,v in j_key.iteritems()])
         s+=" WHERE "
-        s+="AND".join(\
+        s+=_A(\
             ["%s.%s = %%(%s)s " % (r_tbl,v,k)\
                 for k,v in r_key.iteritems()])
 
         return s
 
     def _sqlJoinParent(self, builder, j_builder, outer=False):
-        """ _sqlJoinParent(self, j_builder, builder [, outer= False]) --> (string, string)
+        """ _sqlJoinParent(self, builder, j_builder [, outer= False]) --> (string, string)
 
         Returns a filter and the table where filter has to be applied.
         """
-        tcd, trd= _MotherMappedDeps()
 
         tbl= builder.table_name
         j_tbl= j_builder.table_name
 
-        if j_tbl not in tcd[tbl].keys():
-
-            rel= frozenset([tbl,j_tbl])
+        if not self._isChildOf(j_tbl, tbl):
 
             try:
-                j_tbl= trd[rel]
+                j_tbl= self._searchRelation([tbl, j_tbl])
             except:
-                self.log_int_raise('_sqlJoinParent: invalid join'\
-                        ' table %s!' % ERR_COL(j_tbl))
-
-            self.log_insane( 
-                    "Joining relations %s", 
-                    INF_COL(j_tbl))
-
-        else:
-            self.log_insane( 
-                    "Joining child %s", 
-                    INF_COL(j_tbl))
+                self.log_int_raise('_sqlJoinParent: no way to join |%s| with |%s|.',
+                                    ERR_COL(j_tbl), ERR_COL(tbl))
 
         m_d= self.getChildDeps(j_tbl, builder)
 
@@ -1041,16 +1064,13 @@ class MotherManager:
         else:
             s=" JOIN %s on " % j_tbl
 
-        s+="AND".join(\
-            [" %(tbl)s.%(k)s = %(j_tbl)s.%(v)s " % locals()
-            for k,v in m_d.iteritems()])
+        s+=_A([" %(tbl)s.%(k)s = %(j_tbl)s.%(v)s " % locals() for k,v in m_d.iteritems()])
 
         return s, j_tbl
 
     def _getRelationObject(self, builders):
-        """ _getRelationObject(builders): --> object
+        """ _getRelationObject(builders): --> MotherFly
 
-        object is a MotherOnTheFly object.
         builders is a list of table names.
         """
         
@@ -1069,19 +1089,19 @@ class MotherManager:
         """
 
         father=father or self
-        tbl=father.table_name
+        tbl= father.table_name
 
         # Load father children.....
-        dd= _MotherMappedChildren()
-        d= dd[tbl]
+        d= self._table_children(tbl)
 
         # Get the dependencies for child tbl_name
         try:
-            dd= d[tbl_name]
+            dd= d[tbl_name].copy()
         except:
             self.log_int_raise(\
                 "Child %s not specified for table %s.",\
                     ERR_COL(tbl_name),ERR_COL(tbl))
+
         return dd
 
     def _getDepDict(self, mapping_dict, father= None):
@@ -1099,7 +1119,8 @@ class MotherManager:
                 "Exporting values %s ",
                 str(mapping_dict))
                         
-        fkeys=set(mapping_dict.keys())
+        fkeys= set(mapping_dict.keys())
+
         try:
             d= father.getFields(fkeys)
         except Exception, s:
@@ -1126,7 +1147,7 @@ class MotherManager:
         # Create new dict ...
         new_d= self._getDepDict(d,father)
         # Init child with new_dict
-        child_obj.setFields(new_d,safe_mode=False)
+        child_obj.setFields(new_d, safe_mode= False)
 
         self.log_insane( 
                 "Child object(%s) updated with values %s.",\
@@ -1154,38 +1175,7 @@ class MotherManager:
         return new_obj
         # Return ;)
 
-    def _searchRelation(self, tbls):
-
-        fz= frozenset(tbls)
-
-        trd= _MotherMappedRels()
-        try:
-            r_tbl= trd[fz]
-            self.log_insane(
-                    "Using relation %s to join "
-                    "tables %s", r_tbl,tbls)
-            return r_tbl
-
-        except:
-            pass
-
-        self.log_info(\
-            "Warning: direct relation for set %s "
-            "is not defined: searching subset...",
-            INF_COL(fz))
-
-        rels= [trd[k] for k in trd.keys() if fz< k]
-
-        if len(rels)<>1:
-            self.log_int_raise(
-                "In getRelation: can not get the "  
-                "relation table. There are %s possible "     
-                "choices: %s .", ERR_COL(len(rels)), ERR_COL(
-                ",".join(rels)))
-
-        return rels[0]
-
-       
+      
 #
 ##
 ### Public
@@ -1494,9 +1484,7 @@ class MotherManager:
         # If fields is not specified, fields for the
         # serached table are retrieved 
         if fields== []:
-            tfd= _MotherMappedFields()
-            fields= tfd[b_tbl]
-            del tfd
+            fields= self._table_fields(b_tbl)
 
         self.log_insane("Joining %s and %s with relation %s "\
                         "and fields %s ...",s_tbl,b_tbl,\
@@ -1517,6 +1505,27 @@ class MotherManager:
         mb= MotherBox(builder, qry_filter, MO_LOAD, fields, order, session= self.session)
         return mb
 
+    def relParams(self, related, fields= None):
+        """ relParams(self, related, fields) -> Mother Object
+
+        Returns a Mother Object representing the relation record
+        between `self' and related records.
+        `related' is a list of related records (Mother Objects).
+        if `fields' is not None, only these fields will be loaded."""
+
+        btbls= [b.table_name for b in related]
+
+        # self.table_name is added automatically
+        r_obj = self._getRelationObject(btbls)
+
+        # Let child take his values from fathers...
+        for father in [self] + related:
+            self._fillChildDict(r_obj, father)
+
+        r_obj.load(fields= fields)
+
+        return r_obj
+
     def initChildManager(self, children):
         """initChildManager(self,children) --> None
         
@@ -1542,28 +1551,28 @@ class MotherManager:
         """
 
         def insert_child(builder):
-            def fly(d):
+            def fly_insert_child(d):
                 return self.insertChild(builder, d)
-            return fly
+            return fly_insert_child
 
         def delete_children(builder):
-            def fly(filter= None):
+            def fly_delete_children(filter= None):
                 return self.deleteChildren(builder, filter)
-            return fly
+            return fly_delete_children
 
         def get_children(builder):
-            def fly(fields= None, filter= None, 
+            def fly_get_children(fields= None, filter= None, 
                     jbuilder= None, jfilter= None, 
                     outer= False, order= None):
                 return self.getChildren(builder, fields, 
                         filter, jbuilder, jfilter, 
                         outer, order)
-            return fly
+            return fly_get_children
         
         def update_children(builder):
-            def fly(d_up, filter= None):
+            def fly_update_children(d_up, filter= None):
                 return self.updateChildren(builder, d_up, filter)
-            return fly
+            return fly_update_children
 
 
         camel= DbMother.MotherNaming
@@ -1634,6 +1643,7 @@ class MotherManager:
             - joinRelation
             - dropRelation
             - dropRelations
+            - relParams
 
         The wrappers will be:
 
@@ -1641,29 +1651,37 @@ class MotherManager:
             - joinFoo
             - dropFoo
             - dropMultipleFoo
+            - paramsFoo
 
         NOTE: methods already defined are not overwritten.
         """
 
         def assign_rel(builder):
-            def fly(d, flag= MO_NOA, params= {}):
-                return self.assignRelation([(builder, d, flag)], params)
-            return fly
+            def fly_assign_rel(d, flag= MO_NOA, params= {}):
+                return self.assignRelation([(builder, d, flag)], params)[0]
+            return fly_assign_rel
 
         def drop_rel(builder):
-            def fly(d, flag= MO_NOA, params= {}):
+            def fly_drop_rel(d, flag= MO_NOA, params= {}):
                 return self.dropRelation((builder, d, flag), params)[0]
-            return fly
+            return fly_drop_rel
 
         def drop_rels(builder):
-            def fly(filter= None, jfilter= None, r_tbl= None, flag= MO_NOA):
+            def fly_drop_rels(filter= None, jfilter= None, r_tbl= None, flag= MO_NOA):
                 return self.dropRelations(builder, filter, jfilter, r_tbl, flag)
-            return fly
+            return fly_drop_rels
 
         def join_rels(builder):
-            def fly(fields= [], filter= None, jfilter= None, r_tbl= None, order= None):
+            def fly_join_rels(fields= [], filter= None, jfilter= None, r_tbl= None, order= None):
                 return self.joinRelation(builder, fields, filter, jfilter, r_tbl, order)
-            return fly
+            return fly_join_rels
+
+        def rel_params(builder):
+            def fly_rel_params(d, fields= None):
+                if not isinstance(d, DbMother):
+                    d= builder(d)
+                return self.relParams([d], fields)
+            return fly_rel_params
 
 
         camel= DbMother.MotherNaming
@@ -1717,6 +1735,20 @@ class MotherManager:
 
                 setattr(self, attr_name, attr)
 
+            attr_name= camel(rtbl, 'params')
+            if not hasattr(self, attr_name):
+                attr= rel_params(rel)
+                attr.__doc__=\
+                    " %s(obj [,fields= None]) --> Mother Object\n\n"                \
+                    "Returns a Mother Object that represents the relation\n"        \
+                    "between `self' and the record on the table |%s|, which is\n"   \
+                    "represented by obj (a dict or a Mother Object).\n"             \
+                    "If `fields' is specified, load only these fields."             \
+                    % (attr_name, rtbl)
+
+                setattr(self, attr_name, attr)
+
+
 
 #
 ##
@@ -1794,7 +1826,7 @@ class MotherBox(Psygres):
                 "Mbox: loading with filter=%s",\
                 INF_COL(filter))
 
-        distinct= distinct or ''
+        distinct= distinct and 'DISTINCT' or ''
 
         sel="SELECT %(distinct)s %(what)s FROM %(tbl)s %(filter)s "
 
@@ -1810,12 +1842,12 @@ class MotherBox(Psygres):
             if len(ifields):
                 self.log_int_raise("Invalid Fields for table %s: %s",
                     tbl, ERR_COL(list(ifields)))
-            what= ",".join(['%s.%s ' % (tbl, f) for f in fields])
+            what= _J(['%s.%s' % (tbl, f) for f in fields])
 
         qry= sel % locals()
 
         if order:
-            qry+= " ORDER BY %s" % ','.join(order)
+            qry+= " ORDER BY %s" % _J(order)
 
         self._store= self.mr_query(qry)
 
@@ -1935,5 +1967,3 @@ class MotherBox(Psygres):
             "Obsolete Method MotherBox.getFields(). Use getRecords() instead! "
             "This method will be removed from Mother."))
         return self.getRecords(flag_obj)
-
-
